import open3d as o3d
import numpy as np
from scipy.spatial.transform import Rotation as R
import  copy

def draw_registration_result(source, target, transformation):
    source_temp = copy.deepcopy(source)
    target_temp = copy.deepcopy(target)
    source_temp.paint_uniform_color([1, 0.706, 0])
    target_temp.paint_uniform_color([0, 0.651, 0.929])
    source_temp.transform(transformation)
    o3d.visualization.draw_geometries([source_temp, target_temp])
                                    #   zoom=0.4459,
                                    #   front=[0.9288, -0.2951, -0.2242],
                                    #   lookat=[1.6784, 2.0612, 1.4451],
                                    #   up=[-0.3402, -0.9189, -0.1996])

def transform_pc(pc: np.ndarray, x: float, y: float, z: float, rx: float, ry: float, rz: float) -> np.ndarray:
    rot = R.from_euler('zyx', [rx, ry, rz])
    transform = np.eye(4)
    transform[0, 3] = x
    transform[1, 3] = y
    transform[2, 3] = z

    transform[:3, :3] = rot.as_matrix()

    pc_homo = np.ones((4, pc.shape[0]), dtype=float)
    pc_homo[:3, :] = pc.T

    new_pc = transform @ pc_homo
    return new_pc[:3, :].T

def uniform_sampling(pc: o3d.geometry.PointCloud, every_k_points: int) -> np.ndarray:
    new_pc = pc.uniform_down_sample(every_k_points)
    return new_pc


# read the bunny mesh
mesh_path = "/Users/huang/hcg/tutor/mobile_robotics_ss23/exercise/ex10/bunny/reconstruction/bun_zipper_res2.ply"
mesh = o3d.io.read_triangle_mesh(mesh_path)
mesh.compute_vertex_normals()
o3d.visualization.draw_geometries([mesh])


# convert the mesh to point cloud
pc = np.array(mesh.vertices)
bunny = o3d.geometry.PointCloud()
bunny.points = o3d.utility.Vector3dVector(pc)
# bunny.normals = o3d.utility.Vector3dVector(mesh.triangle_normals)
# bunny.colors = o3d.utility.Vector3dVector(mesh.triangle_normals)
bunny.estimate_normals()
bunny.normalize_normals()
# bunny = uniform_sampling(bunny, 10)


# create a source point cloud
bunny_2 = copy.deepcopy(bunny)
pc_2 = transform_pc(np.asarray(bunny.points), 0, 0, 0.1, 0.4, 0, 0)
bunny_2.points = o3d.utility.Vector3dVector(pc_2)



draw_registration_result(bunny_2, bunny, np.eye(4))

threshold = 0.2
trans_init = np.eye(4)
for i in range(10):
    reg_p2p = o3d.pipelines.registration.registration_icp(
        bunny_2, bunny, threshold, trans_init,
        # o3d.pipelines.registration.TransformationEstimationPointToPoint(),
        o3d.pipelines.registration.TransformationEstimationPointToPlane(),
        o3d.pipelines.registration.ICPConvergenceCriteria(max_iteration=1))
    evaluation = o3d.pipelines.registration.evaluate_registration(
        bunny_2, bunny, threshold, reg_p2p.transformation)
    print(f"Iteration {i+1}: ", evaluation)
    bunny_2.transform(reg_p2p.transformation)
    draw_registration_result(bunny_2, bunny, reg_p2p.transformation)
