import numpy as np
import math
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist
import time

from misc_tools import *

def closest_point_matching(X, P):
  """
  Performs closest point matching of two point sets.
  
  Arguments:
  X -- reference point set (3, N)
  P -- point set to be matched with the reference (3, M)
  
  Output:
  P_matched -- reordered P, so that the elements in P match the elements in X
  """

  P_matched = P

  # TODO: implement

  X_np = np.array(X)
  P_np = np.array(P)
  dist_mat = cdist(X_np.T, P_np.T, 'euclidean')
  P_matched = np.zeros_like(P)
  for k in range(dist_mat.shape[1]):
    min_r, min_c = np.unravel_index(np.argmin(dist_mat), dist_mat.shape)
    P_matched[:, min_r] = P[:, min_c]
    dist_mat[min_r, :] = np.Inf
    dist_mat[:, min_c] = np.Inf


  return P_matched
  
def icp(X, P, do_matching):
  
  P0 = P  
  for i in range(10):  
    # calculate RMSE
    rmse = 0
    for j in range(P.shape[1]):
      rmse += math.pow(P[0,j] - X[0,j], 2) + math.pow(P[1,j] - X[1,j], 2)
    rmse = math.sqrt(rmse / P.shape[1])

    # print and plot
    print("Iteration:", i, " RMSE:", rmse)
    plot_icp(X, P, P0, i, rmse)
    
    # data association
    if do_matching:
      P = closest_point_matching(X, P)
    plot_icp(X, P, P0, i, rmse)
    
    # substract center of mass
    mx = np.transpose([np.mean(X, 1)])
    mp = np.transpose([np.mean(P, 1)])
    X_prime = X - mx
    P_prime = P - mp
    
    # singular value decomposition
    W = np.dot(X_prime, P_prime.T)
    U, s, V = np.linalg.svd(W)
    
    # calculate rotation and translation
    val = np.linalg.det(U @ V.T)
    print(val)
    if val < 0:
      V[:, -1] = -V[:, -1]
    R = np.dot(U, V.T)
    t = mx - np.dot(R, mp)
    print(R, t)
    
    # apply transformation
    plot_icp(X, P, P0, i, rmse)
    P = np.dot(R, P) + t
    # P = P + t
    plot_icp(X, P, P0, i, rmse)
    plt.waitforbuttonpress()

  return
    
def main():
  
  X, P1, P2, P3, P4 = generate_data()
  
  # icp(X, P1, False)
  # icp(X, P2, False)
  # icp(X, P3, True)
  icp(X, P4, True)


  plt.waitforbuttonpress()
    
if __name__ == "__main__":
  main()
